﻿using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace TSLab.Script.Helpers
{
    /// <summary>
    /// \~english Double values list functions \~russian Список функций значений чисел
    /// </summary>
    public static class Series
    {
        /// <summary>
        /// \~english Calculate highest double value \~russian Рассчитать максимальное значение
        /// </summary>
        /// <param name="period">\~english calculate period \~russian Период расчета</param>
        /// <param name="candles">\~english latest calculation bar \~russian Последний бар в расчете</param>
        /// <returns>\~english highest double value \~russian Возвращает максимальное значение</returns>
        public static IList<double> Highest(IList<double> candles, int period)
        {
            var count = candles.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                res[i] = Indicators.Highest(candles, i, period);
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate lowest double value \~russian Рассчитать минимальное значение
        /// </summary>
        /// <param name="period">\~english calculate period \~russian Период расчета</param>
        /// <param name="candles">\~english latest calculation bar \~russian Последний бар в расчете</param>
        /// <returns>\~english Lowest double value \~russian Возвращает минимальное значение</returns>
        public static IList<double> Lowest(IList<double> candles, int period)
        {
            var count = candles.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                res[i] = Indicators.Lowest(candles, i, period);
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate new double values list with minimums of source lists 
        /// \~russian Рассчитать новый список значений чисел с минимумами входящих списков
        /// </summary>
        /// <param name="src1">\~english the first source list \~russian Первый входящий список</param>
        /// <param name="src2">\~english the second source list \~russian Второй входящий список</param>
        /// <returns>\~english list with minimum values \~russian Возвращает список с минимальными значениями</returns>
        public static IList<double> Min(IList<double> src1, IList<double> src2)
        {
            int count = Math.Max(src1.Count, src2.Count);
            var dst = new double[count];
            for (int i = 0; i < count; i++)
            {
                var s1 = src1.Count == 0 ? 0 : src1[Math.Min(i, src1.Count - 1)];
                var s2 = src2.Count == 0 ? 0 : src2[Math.Min(i, src2.Count - 1)];
                dst[i] = Math.Min(s1, s2);
            }
            return dst;
        }

        /// <summary>
        /// \~english Calculate new double values list with maximums of source lists 
        /// \~russian Рассчитать новый список значений чисел с максимумами входящих списков
        /// </summary>
        /// <param name="src1">\~english the first source list \~russian Первый входящий список</param>
        /// <param name="src2">\~english the second source list \~russian Второй входящий список</param>
        /// <returns>\~english list with maximum values \~russian Возвращает список с максимальными значениями</returns>
        public static IList<double> Max(IList<double> src1, IList<double> src2)
        {
            int count = Math.Max(src1.Count, src2.Count);
            var dst = new double[count];
            for (int i = 0; i < count; i++)
            {
                var s1 = src1.Count == 0 ? 0 : src1[Math.Min(i, src1.Count - 1)];
                var s2 = src2.Count == 0 ? 0 : src2[Math.Min(i, src2.Count - 1)];
                dst[i] = Math.Max(s1, s2);
            }
            return dst;
        }

        /// <summary>
        /// \~english Calculate new double values list with summ of source lists 
        /// \~russian Рассчитать новый список значений чисел с суммой входящих списков
        /// </summary>
        /// <param name="src1">\~english the first source list \~russian Первый входящий список</param>
        /// <param name="src2">\~english the second source list \~russian Второй входящий список</param>
        /// <returns>\~english list with minimum values \~russian Возвращает список с минимальными значениями</returns>
        public static IList<double> Add(IList<double> src1, IList<double> src2)
        {
            int count = Math.Max(src1.Count, src2.Count);
            var dst = new double[count];
            for (int i = 0; i < count; i++)
            {
                var s1 = src1.Count == 0 ? 0 : src1[Math.Min(i, src1.Count - 1)];
                var s2 = src2.Count == 0 ? 0 : src2[Math.Min(i, src2.Count - 1)];
                dst[i] = s1 + s2;
            }
            return dst;
        }

        /// <summary>
        /// \~english Calculate new double values list with subtraction of source lists 
        /// \~russian Рассчитать новый список значений чисел с вычитанием входящих списков
        /// </summary>
        /// <param name="src1">\~english the first source list \~russian Первый входящий список</param>
        /// <param name="src2">\~english the second source list \~russian Второй входящий список</param>
        /// <returns>\~english list with minimum values \~russian Возвращает список с минимальными значениями</returns>
        public static IList<double> Sub(IList<double> src1, IList<double> src2)
        {
            int count = Math.Max(src1.Count, src2.Count);
            var dst = new double[count];
            for (int i = 0; i < count; i++)
            {
                var s1 = src1.Count == 0 ? 0 : src1[Math.Min(i, src1.Count - 1)];
                var s2 = src2.Count == 0 ? 0 : src2[Math.Min(i, src2.Count - 1)];
                dst[i] =s1 - s2;
            }
            return dst;
        }

        /// <summary>
        /// \~english Calculate new int values list with summ of source lists 
        /// \~russian Рассчитать значения нового числового списка с суммой входящих списков
        /// </summary>
        /// <param name="src1">\~english the first source list \~russian Первый входящий список</param>
        /// <param name="src2">\~english the second source list \~russian Второй входящий список</param>
        /// <returns>\~english list with minimum values \~russian Возвращает список с минимальными значениями</returns>
        public static IList<int> Add(IList<int> src1, IList<int> src2)
        {
            int count = Math.Max(src1.Count, src2.Count);
            IList<int> dst = new List<int>(count);
            for (int i = 0; i < count; i++)
            {
                var s1 = src1.Count == 0 ? 0 : src1[Math.Min(i, src1.Count - 1)];
                var s2 = src2.Count == 0 ? 0 : src2[Math.Min(i, src2.Count - 1)];
                dst.Add(s1 + s2);
            }
            return dst;
        }

        /// <summary>
        /// \~english Calculate the list of the Median prices
        /// \~russian Расчитать список средних цен
        /// </summary>
        /// <param name="candles">\~english The source candles list \~russian Список свечей</param>
        /// <returns>\~english Median prices list \~russian Список средних цен</returns>
        public static IList<double> MedianPrice(IList<Bar> candles)
        {
            var count = candles.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                res[i] = (candles[i].High + candles[i].Low)/2;
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate True Range for the all candles \~russian Рассчитать TR (истинный диапазон) для всех баров
        /// </summary>
        /// <param name="candles">\~english source candles list \~russian Список баров</param>
        /// <returns>\~english true ranges list \~russian Список TR</returns>
        public static IList<double> TrueRange(IList<Bar> candles)
        {
            var count = candles.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                res[i] = Indicators.TrueRange(candles, i);
            }
            return res;
        }

        /// <summary>
        /// \~english ATR helper \~russian Помощник ATR (средний истинный диапазон)
        /// </summary>
        /// <param name="candles"></param>
        /// <param name="start"></param>
        /// <param name="period"></param>
        /// <param name="prevAtr"></param>
        /// <returns></returns>
        public static double AverageTrueRange(IList<Bar> candles, int start, int period, double prevAtr)
        {
            if (start - period > 0)
            {
                return (prevAtr*(period - 1) + Indicators.TrueRange(candles, start))/period;
            }
            double atr = prevAtr * start;
            atr += Indicators.TrueRange(candles, start);
            return atr/(start + 1);
        }

        /// <summary>
        /// \~english Calculate Average True Range for the all candles \~russian Рассчитать ATR (средний истинный диапазон) для всех свечей
        /// </summary>
        /// <param name="candles">\~english source candles list \~russian Список свечей</param>
        /// <param name="period">\~english ATR period \~russian Период ATR</param>
        /// <returns>\~english ATRs list \~russian Список ATR</returns>
        public static IList<double> AverageTrueRange(IList<Bar> candles, int period)
        {
            double atr = 0;
            var count = candles.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                atr = res[i] = AverageTrueRange(candles, i, period, atr);
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate SMA for source list \~russian Рассчитать SMA (простое скользящее среднее) для входящего списка
        /// </summary>
        /// <param name="period">\~english SMA period \~russian Период SMA</param>
        /// <param name="candles">\~english source candles list \~russian Входящий список</param>
        /// <returns>\~english SMAs list \~russian Список SMA</returns>
        public static IList<double> SMA(IList<double> candles, int period)
        {
            var count = candles.Count;
            var res = new double[count];

            if(period < 4)
            {
                for (int i = 0; i < count; i++)
                {
                    res[i] = Indicators.SMA(candles, i, period);
                }
            }
            else
            {
                var min = Math.Min(count, period);
                double summ = 0;
                for (int i = 0; i < min; i++)
                {
                    summ += candles[i];
                    res[i] = summ / (i + 1);
                }
                for (int i = min; i < count; i++)
                {
                    double close = candles[i];
                    double prevClose = candles[i - period];
                    double prevSMA = res[i - 1];
                    res[i] = prevSMA + (close - prevClose) / period;
                }
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate Summ for same values\~russian Рассчитать сумму за несколько значений
        /// </summary>
        /// <param name="period">\~english the period \~russian Период </param>
        /// <param name="candles">\~english source candles list \~russian Входящий список</param>
        /// <returns>\~english Summs list \~russian Список сумм</returns>
        public static IList<double> SummFor(IList<double> candles, int period)
        {
            var count = candles.Count;
            var res = new double[count];
            var min = Math.Min(count, period);
            double summ = 0;
            for (int i = 0; i < min; i++)
            {
                summ += candles[i];
                res[i] = summ;
            }
            for (int i = min; i < count; i++)
            {
                double close = candles[i];
                double prevClose = candles[i - period];
                double prevSumm = res[i - 1];
                res[i] = prevSumm + (close - prevClose);
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate StDev for source list \~russian Рассчитать StDev (Стандартное Отклонение) для входящего списка
        /// </summary>
        /// <param name="period">\~english StDev period \~russian Период StDev</param>
        /// <param name="candles">\~english source candles list \~russian Входящий список</param>
        /// <returns>\~english StDev list \~russian Список StDev</returns>
        public static IList<double> StDev(IList<double> candles, int period)
        {
            var count = candles.Count;
            var res = new double[count];
            var sma = SMA(candles, period);
            for (int i = 0; i < count; i++)
            {
                res[i] = Indicators.StDev(candles, sma, i, period);
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate moved SMA for source list \~russian Рассчитать сдвинутое SMA для входящего списка
        /// </summary>
        /// <param name="candles">\~english source candles list \~russian Входящий список</param>
        /// <param name="period">\~english SMA period \~russian Период SMA</param>
        /// <param name="shift">\~english SMA shift \~russian Сдвиг SMA</param>
        /// <returns>\~english moved SMAs list \~russian Список SMA</returns>
        public static IList<double> SMMA(IList<double> candles, int period, int shift)
        {
            var list = SMA(candles, period);
            return Shift(list, shift);
        }

        /// <summary>
        /// Shift source list to future
        /// </summary>
        /// <param name="input">the source list</param>
        /// <param name="shift">the shifting period</param>
        /// <returns>the shifted list</returns>
        public static IList<double> Shift(IList<double> input, int shift)
        {
            var list= new List<double>(input);
            for (int i = 0; list.Count > 0 && i < shift; i++)
            {
                list.RemoveAt(list.Count - 1);
                list.Insert(0, list[1]);
            }
            return list;
        }

        /// <summary>
        /// \~english Calculate EMA for source list \~russian Рассчитать EMA (экспоненциально сглаженное скользящее среднее) для входящего списка 
        /// </summary>
        /// <param name="period">\~english EMA period \~russian Период EMA</param>
        /// <param name="candles">\~english source candles list \~russian Входящий список</param>
        /// <returns>\~english EMAs list \~russian Список EMA</returns>
        public static IList<double> EMA(IList<double> candles, int period)
        {
            var count = candles.Count;
            var emas = new double[count];
            
            var min = Math.Min(count, period);
            double summ = 0;
            for (int i = 0; i < min; i++)
            {
                summ += candles[i];
                emas[i] = summ / (i+1);
            }

            double k = 2.0/(1.0 + period);
            for (int i = min; i < count; i++)
            {
                double close = candles[i];
                double prevEMA = emas[i - 1];
                emas[i] = k*(close - prevEMA) + prevEMA;
                Debug.Assert(!double.IsNaN(emas[i]));
            }
            return emas;
        }

        /// <summary>
        /// \~english Calculate RSI for source list \~russian Рассчитать RSI для входящего списка 
        /// </summary>
        /// <param name="period">\~english RSI period \~russian Период RSI</param>
        /// <param name="candles">\~english source candles list \~russian Входящий список баров</param>
        /// <returns>\~english RSI list \~russian Список RSI</returns>
        public static IList<double> RSI(IList<double> candles, int period)
        {
            var count = candles.Count;
            var res = new double[count];
            if (count > 0)
            {
                var u = new double[count];
                var d = new double[count];
                u[0] = 0;
                d[0] = 0;
                for (int i = 1; i < count; i++)
                {
                    double tu = 0, td = 0;
                    if (candles[i - 1] < candles[i])
                    {
                        tu = candles[i] - candles[i - 1];
                    }
                    else if (candles[i - 1] > candles[i])
                    {
                        td = candles[i - 1] - candles[i];
                    }
                    u[i] = tu;
                    d[i] = td;
                }
                var eu = EMA(u, period);
                var ed = EMA(d, period);
                for (int i = 0; i < count; i++)
                {
                    if(ed[i] == 0.0)
                    {
                        res[i] = 100;
                    }
                    else if (eu[i] / ed[i] == 1.0)
                    {
                        res[i] = 0;
                    }
                    else
                    {
                        res[i] = 100 - 100 / (1 + eu[i] / ed[i]);
                    }
                }
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate Cuttler RSI for source list \~russian Рассчитать Cuttler RSI для входящего списка 
        /// </summary>
        /// <param name="period">\~english RSI period \~russian Период RSI</param>
        /// <param name="candles">\~english source candles list \~russian Входящий список баров</param>
        /// <returns>\~english RSI list \~russian Список RSI</returns>
        public static IList<double> CuttlerRSI(IList<double> candles, int period)
        {
            var count = candles.Count;
            var res = new double[count];
            if (count > 0)
            {
                var u = new double[count];
                var d = new double[count];
                u[0] = 0;
                d[0] = 0;
                for (int i = 1; i < count; i++)
                {
                    double tu = 0, td = 0;
                    if (candles[i - 1] < candles[i])
                    {
                        tu = candles[i] - candles[i - 1];
                    }
                    else if (candles[i - 1] > candles[i])
                    {
                        td = candles[i - 1] - candles[i];
                    }
                    u[i] = tu;
                    d[i] = td;
                }
                var eu = SMA(u, period);
                var ed = SMA(d, period);
                for (int i = 0; i < count; i++)
                {
                    if (ed[i] == 0.0)
                    {
                        res[i] = 100;
                    }
                    else if (eu[i] / ed[i] == 1.0)
                    {
                        res[i] = 0;
                    }
                    else
                    {
                        res[i] = 100 - 100 / (1 + eu[i] / ed[i]);
                    }
                }
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate the "Typical Price" indicator \~russian Рассчитать индикатор TP (Typical Price, типичная цена)
        /// </summary>
        /// <param name="candles">\~english source candles list \~russian Входящий список баров</param>
        /// <returns>\~english calculated values \~russian Возвращает рассчитанные значения</returns>
        public static IList<double> TypicalPrice(IList<Bar> candles)
        {
            var count = candles.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                res[i] = (candles[i].High + candles[i].Low + candles[i].Close)/3;
            }
            return res;
        }

        private static double MA(IList<double> candles, int curbar, int period, double sma)
        {
            int start = curbar - period + 1;
            if (start < 0)
            {
                start = 0;
            }
            period = curbar - start + 1;
            double sum = 0;
            var count = candles.Count;
            while (start <= curbar && start < count)
            {
                sum += Math.Abs(candles[start++] - sma);
            }
            sum /= Math.Max(1, period);
            return sum;
        }

        /// <summary>
        /// \~english Calculate CCI indicator \~russian Рассчитать индикатор CCI (индекс товарного канала)
        /// </summary>
        /// <param name="candles">\~english source candles list \~russian Входящий список баров</param>
        /// <param name="period">\~english period of MA \~russian Период MA</param>
        /// <returns>\~english calculated values \~russian Возвращает рассчитанные значения</returns>
        public static IList<double> CCI(IList<Bar> candles, int period)
        {
            var tr = TypicalPrice(candles);
            var smaTr = SMA(tr, period);
            var count = candles.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                res[i] = tr[i] - smaTr[i];
            }
            for (int i = 0; i < count; i++)
            {
                var d = (MA(tr, i, period, smaTr[i])*0.015);
                res[i] = d == 0.0 ? 0.0 : res[i]/d;
            }
            return res;
        }

        /// <summary>
        /// \~english Calculate the Bollinger Bands \~russian Рассчитать границы Боллинджера
        /// </summary>
        /// <param name="closes">\~english the list with the source data (Close values, for example) 
        /// \~russian Входящий список данных (значения закрытий, например)</param>
        /// <param name="period">\~english SMA period \~russian Период SMA</param>
        /// <param name="coef">\~english BB coef \~russian Коэффициент границ Боллинджера</param>
        /// <param name="isTopLine">\~english BB+ if true \~russian Верхняя граница Боллинджера если True</param>
        /// <returns>\~english calculated values \~russian Возвращает рассчитанные значения</returns>
        public static IList<double> BollingerBands(IList<double> closes, int period, double coef, bool isTopLine)
        {
            IList<double> sma = SMA(closes, period);
            return BollingerBands(closes, sma, period, coef, isTopLine);
        }

        /// <summary>
        /// \~english Advanced of Bollinger Bands \~russian Продвинутые границы Боллинджера
        /// </summary>
        /// <param name="closes">\~english the list with the source data (Close values, for example) 
        /// \~russian Входящий список данных (значения закрытий, например)</param>
        /// <param name="sma">\~english the list with calculated base (for example, SMA or EMA) 
        /// \~russian Входящий список рассчитанных базовых параметров (например, SMA или EMA)</param>
        /// <param name="period">\~english second SMA period \~russian Второй период SMA</param>
        /// <param name="coef">\~english BB coef \~russian Коэффициент границ Боллинджера</param>
        /// <param name="isTopLine">\~english BB+ if true \~russian Верхняя граница Боллинджера если True</param>
        /// <returns>\~english calculated values \~russian Возвращает рассчитанные значения</returns>
        public static IList<double> BollingerBands(IList<double> closes, IList<double> sma,
                                                   int period, double coef, bool isTopLine)
        {
            double sign = isTopLine ? 1 : -1;
            var count = closes.Count;
            var res = new double[count];
            for (int i = 0; i < count; i++)
            {
                var stdev = Indicators.StDev(closes, sma, i, period);
                res[i] = sma[i] + sign*coef*stdev;
            }
            return res;
        }
    }
}